@@ -8,15 +8,13 @@ import re
 import sys
 import os
 
-#from xml.parsers.expat import ParserCreate
-#from xml.dom.minidom import Document, Text
 from lxml import etree
 
 from namespace import create_namespace
 from iterator import create_iterator
 from device import Device
-from change import CHANGE_STATE, ChangeConditionFailure, NoCommands
-from provisioner import UserBailout
+from change import ChangeConditionFailure, NoCommands
+from provisioner import UserBailout, NoTargetsError
 import util
 
 from twisted.internet import defer, reactor
@@ -39,6 +37,8 @@ class ConfigLoader:
 
         self.doc = None
 
+        self.global_namespace = {}
+
         self.provisioners = {}
         self.devices = {}
         self.change_templates = {}
@@ -50,45 +50,41 @@ class ConfigLoader:
         self.change_failure = []
         self.backout_success = []
         self.backout_failure = []
+        self.change_skipped = []
 
         log.debug("created ConfigLoader")
 
-    def load_config(self, options=None, devices=[]):
+    def load_config(self, options, devices=[]):
 
         self.options = options
 
         d = defer.succeed(None)
 
         if options.configfile:
-            d.addCallback(self.parse, options.configfile)
+            d.addCallback(self.parse, options.configfile, devices)
             pass
 
-        # Add devices specified on the commandline to the active configuration
-        if len(devices) != 0:
-            for dev in devices:
-                # If a device given on the commandline is not present in the
-                # configuration file, add it.
-                if dev not in self.devices.keys():
-                    log.info("Adding commandline specified device: '%s'" % dev)
-                    self.devices[dev] = Device(dev)
-                    pass
+        # skip changes marked to be skipped on commandline
+        for change in self.changes.values():
+            if change.name in options.skip_changes:
+                log.info("Skipping change: %s", change.name)
+                change.state('skipped')
+                self.change_complete(None, change)
                 pass
-                # If devices other than those specified on the commandline
-                # ARE present in the config file, remove the extra ones.
-                # This means you can run a configuration on a subset of devices
-                # by specifying the device name on the commandline.
-
-            for dev in self.devices.keys():
-                if dev not in devices:
-                    log.info("Not using configured device '%s' (not on commandline)", dev)
-                    del self.devices[dev]
-                    pass
+            pass
+        
+        # If only-changes were marked, mark all other changes as skipped
+        for change in self.changes.values():
+            if len(options.only_changes) > 0 and change.name not in options.only_changes:
+                log.info("Skipping change: %s", change.name)                
+                change.state('skipped')        
+                self.change_complete(None, change)
                 pass
             pass
 
         return d
             
-    def parse(self, ignored, configfile):
+    def parse(self, ignored, configfile, devicenames):
         """
         Parse my configuration file.
         """
@@ -112,7 +108,8 @@ class ConfigLoader:
             nsnode = self.tree.xpath('/config/namespace')[0]
             self.add_namespace(None, nsnode)
         except IndexError:
-            self.global_namespace = {}
+            # use default global_namespace defined in __init__
+            pass
 
         d = defer.succeed(None)
 
@@ -121,56 +118,105 @@ class ConfigLoader:
             'iterator',
             'provisioner',
             'device',
+            ]:
+            self.parse_nodes(d, nodename)
+            pass
+
+        # Add devices specified on the commandline to the active configuration
+        d.addCallback(self.add_commandline_devices, devicenames)
+        
+        # Now process the changes
+        for nodename in [
             'changetemplate',
             'change',
-#            'dependencies',
             ]:
-
-            for node in self.tree.findall(nodename):
-                # Find the function to call based on the name of
-                # the object by using dynamic lookup
-                # This will call 'add_iterator' for an iterator node, for example.
-                funcname = 'add_%s' % nodename
-                func = getattr(self, funcname)
-                d.addCallback(func, node)
-                pass
+            self.parse_nodes(d, nodename)
             pass
         
         d.addCallback(self.set_prereqs)
         return d
 
+    def parse_nodes(self, d, nodename):
+        """
+        Parse all nodes with a given name
+        @param d: a L{defer.Deferred} to callback when finished
+        """
+        for node in self.tree.findall(nodename):
+            # Find the function to call based on the name of
+            # the object by using dynamic lookup
+            # This will call 'add_iterator' for an iterator node, for example.
+            log.debug("Adding a '%s' node", nodename)
+            funcname = 'add_%s' % nodename
+            func = getattr(self, funcname)
+            d.addCallback(func, node)
+            pass
+        pass
+
+    def add_commandline_devices(self, ignored, devicenames):
+        """
+        Add devices specified on the commandline to the active config
+        """
+
+        if len(devicenames) != 0:
+            # If a device given on the commandline is not present in the
+            # configuration file, add it.
+            for dev in devicenames:
+                if dev not in self.devices.keys():
+                    log.info("Adding commandline specified device: '%s'" % dev)
+                    self.devices[dev] = Device(dev)
+                    pass
+                pass
+            # If devices other than those specified on the commandline
+            # ARE present in the config file, remove the extra ones.
+            # This means you can run a configuration on a subset of devices
+            # by specifying the device name on the commandline.
+            for dev in self.devices.keys():
+                if dev not in devicenames:
+                    log.info("Not using configured device '%s' (not on commandline)", dev)
+                    del self.devices[dev]
+                    pass
+                pass
+            pass
+
     def set_prereqs(self, ignored):
+        log.debug("Setting prereqs...")
 
         # Find all the prereq definitions
-        prereqs = self.tree.findall('prereq')
+        prereqs = self.tree.xpath('descendant-or-self::*/prereq')
+        
+        log.debug("found prereq nodes: %s", prereqs)
         for node in prereqs:
             prereq_changename = node.text
 
             try:
-                prereq_for = node.attrib['for']
+                prereq_for_name = node.attrib['for']
             except KeyError:
                 # See if this prereq node is within a change node
-                prereq_for = node.xpath("parent::*/change")[0].attrib['name']
-                log.debug("Using change name of '%s'", prereq_for)
+                try:
+                    prereq_for_name = node.xpath("parent::*")[0].attrib['name']
+                except KeyError:
+                    log.error("prerequisite change '%s' is not defined.", prereq_for_name)
                 
             # find the change
             try:
                 prereq_change = self.changes[prereq_changename]
+                log.debug("prereq_changename: %s", prereq_changename)
             except KeyError:
                 log.error("Cannot find change '%s' as prereq for change '%s'", prereq_changename, prereq_for)
                 raise
 
             try:
-                change_for = self.changes[prereq_for]
+                change_for = self.changes[prereq_for_name]
             except KeyError:
-                log.error("Cannot find change '%s' with prereq of '%s'", prereq_for, prereq_changename)
+                log.error("Cannot find change '%s' with prereq of '%s'", prereq_for_name, prereq_changename)
                 raise
 
             # add change as prereq
-            change_for.pre_requisites.append(prereq_change)
-            log.debug("Added change '%s' as prereq for change '%s'", prereq_changename, prereq_for)
+            change_for.add_prereq(prereq_change)
+            log.debug("Added change '%s' as prereq for change '%s'", prereq_changename, prereq_for_name)
 
         # once the parse has completed, do some post-parse setup
+        # FIXME: This does nothing. Remove it.
         self.process_dependencies()
 
     def add_iterator(self, ignored, node):
@@ -222,9 +268,20 @@ class ConfigLoader:
         prov_module = self.import_module(module_name)
         log.debug("prov_module is: %s", prov_module)
         klass = getattr( prov_module, prov_klass )
+
+        if self.options.sessionlog is not None:
+            sessionlog = open(self.options.sessionlog, "w")
+        else:
+            # always create a session log, named after the xml config file
+            filename = "%s.log" % os.path.splitext(self.options.configfile)[0]
+            sessionlog = open(filename, "w")
+            #sessionlog = None
+        
         try:
-            provisioner = klass(prov_name, authoritarian=self.options.authoritarian,
+            provisioner = klass(prov_name,
+                                authoritarian=self.options.authoritarian,
                                 autobackout=self.options.autobackout,
+                                sessionlog=sessionlog,
                                 **kwargs)
         except TypeError, e:
             log.error("Incorrect parameter supplied for provisioner of type '%s'", prov_klass)
@@ -234,11 +291,19 @@ class ConfigLoader:
 
         # If there are further attributes for the provisioner,
         # handle them with the provisioner itself
-        for subnode in node.xpath('*'):
+        nodelist = node.xpath('*')
+        taglist = [ x.tag for x in nodelist ]
+        # Make sure all the required nodes exist
+        for tag in provisioner.get_required_tags():
+            if tag not in taglist:
+                raise ValueError("Required tag '%s' not found in '%s' definition." % (tag, provisioner.name))
+
+        for subnode in nodelist:
             provisioner.parse_config_node(subnode)
             pass
 
         self.provisioners[prov_name] = provisioner
+        return provisioner
 
     def add_device(self, ignored, node):
         """
@@ -256,6 +321,7 @@ class ConfigLoader:
 
         log.debug("created device: %s", device)
         self.devices[device_name] = device
+        return device
 
     def add_changetemplate(self, ignored, node):
         """
@@ -334,13 +400,16 @@ class ConfigLoader:
                     pass
                 pass
             pass
+
         # If, after all that, we can't find a provisioner for the change, bail.
         if change.provisioner is None:
             raise ValueError("No compatible provisioners exist for change '%s'" % change.name)
         
         self.changes[change.name] = change
+        log.debug("Adding change '%s' to pending changes: %s", change.name, self.pending_changes)
         self.pending_changes.append( change )
-
+        return change
+        
     def create_change_object(self, node):
         """
         Create a change object from a node.
@@ -355,6 +424,10 @@ class ConfigLoader:
             log.error("Change definition has no name")
             raise
 
+        # Check to see if the change name already exists:
+        if change_name in self.changes:
+            raise KeyError("Change '%s' already exists. Duplicate change name defined.", change_name)
+        
         try:
             change_klass = node.attrib['type']
         except KeyError:
@@ -411,12 +484,14 @@ class ConfigLoader:
         for subnode in node.xpath('*'):
             # a target device for the change
             if subnode.tag == 'target':
-                targetname = subnode.text
+                targetname = subnode.attrib['name']
 
                 log.debug("-- Found targetname: %s", targetname)
 
                 # Parse special targetname ALL_TARGETS
                 if targetname == 'ALL_TARGETS':
+                    if len(self.devices) == 0:
+                        raise ValueError("No devices defined. ALL_TARGETS specifies nothing.")
                     change.devices.extend( self.devices.values() )
                     devicenames = ','.join( [ '%s' % x for x in self.devices.keys() ] )
                     log.debug("added target devices '%s' for change '%s'", devicenames, change)
@@ -429,8 +504,8 @@ class ConfigLoader:
                         raise KeyError("Device '%s' not defined" % targetname)
                     change.devices.append(dev)
                     log.debug("added target device '%s' for change '%s'", dev, change)
+                    pass
                 pass
-
             elif subnode.tag == 'depends':
                 changename = subnode.attrib['on']
                 try:
@@ -455,7 +530,11 @@ class ConfigLoader:
             #
             else:
                 method_name = 'parse_%s' % str(subnode.tag)
-                method = getattr(change, method_name)
+                try:
+                    method = getattr(change, method_name)
+                except AttributeError, e:
+                    raise AttributeError("'%s' is not a valid sub-element of %s" % (subnode.tag, change))
+                
                 method(subnode)
                 pass
             pass
@@ -489,7 +568,7 @@ class ConfigLoader:
         log.debug("Checking for 'onfail' attribute...")
         try:
             onfail = node.attrib['onfail']
-            log.debug("onfail node found.")
+            log.debug("onfail attrib found.")
             if onfail == 'continue':
                 change.on_fail_continue = True
                 log.debug("Processing will continue if this change fails")
@@ -516,6 +595,20 @@ class ConfigLoader:
         except KeyError:
             pass
 
+    def set_change_noop(self, change, node):
+        """
+        Mark a change as a 'No Op' if it has the noop attribute set.
+        If it's not set to anything, this is taken as meaning 'true'.
+        If it's absent, then it
+        """
+        try:
+            onfail = node.attrib['noop']
+            log.debug("noop attrib found")
+            if not ( noop.lower().startswith('n') or noop.lower().startswith('f') ):
+                change.noop = True
+        except KeyError:
+            pass
+
     def add_namespace(self, ignored, node, item=None):
         """
         Add a namespace, defined by nsnode, to an item (such as a change)
@@ -534,6 +627,8 @@ class ConfigLoader:
             ns.parent = self.global_namespace
             item.namespace = ns
 
+        return ns
+
     def get_available_changes(self):
         """
         This returns a list of pending changes that can be performed,
@@ -542,43 +637,69 @@ class ConfigLoader:
         """
         changelist = []
         log.debug("pending changes: %s", self.pending_changes)
-        for change in self.pending_changes:
-            log.debug("testing change %s", change)
-            if len(change.pre_requisites) == 0:
-                changelist.append(change)
-                log.debug("change '%s' has no pre-reqs. Adding to execution queue.", change)
-                continue
 
-            else:
-                all_prereqs = True
-                for prereq in change.pre_requisites:
-                    log.debug("change '%s' has a pre-req of '%s'", change, prereq)
-                    if prereq not in self.change_success:
+        # Am I in backout mode?
+        if self.options.backout:
+            log.info("In backout mode. Doing changes in reverse.")
+            # Find all changes that don't have children,
+            # or that have children that have already been run
+            for change in self.pending_changes:
+                if not change.has_children():
+                    changelist.append(change)
+                    log.debug("change '%s' has no children. Adding to backout change queue.", change)
+                    continue
 
-                        # If we're in backout mode, a prereq in 'backout_ok' state
-                        # is treated as complete.
-                        if prereq.state in [ CHANGE_STATE['backout_ok'], ]:
-                            continue
-
-                        # If a prereq has failed, doesn't need to retry, and is
-                        # marked as 'onfail:continue', then we treat it as if
-                        # this prereq has been met.
-                        elif prereq.state not in [ CHANGE_STATE['pending'],
-                                                 CHANGE_STATE['retry'],
-                                                 ] and prereq.on_fail_continue:
-                            log.debug("prereq failed, but is marked onfail:continue.")
-                            continue
-
-                        log.debug("pre-req '%s' not complete yet.", prereq)
-                        all_prereqs = False
+                # check all my children have backed out ok
+                all_children = True
+                for child_change in change.children:
+                    if child_change.state() not in ['backout_ok',
+                                                    'skipped',
+                                                    'success',
+                                                    ]:
+                        all_children = False
                         break
                     pass
-
-                if all_prereqs:
-                    log.debug("All pre-reqs for '%s' have completed. Adding to execution queue.", change)
+                if all_children:
                     changelist.append(change)
+                
+        else:
+            # FIXME: This looks too complicated. Can it can be simplified?
+            for change in self.pending_changes:
+                log.debug("testing change %s", change)
+
+                if len(change.pre_requisites) == 0:
+                    changelist.append(change)
+                    log.debug("change '%s' has no pre-reqs. Adding to execution queue.", change)
+                    continue
+
                 else:
-                    log.debug("change '%s' has pending pre-reqs", change)
+                    all_prereqs = True
+                    for prereq in change.pre_requisites:
+                        log.debug("change '%s' has a pre-req of '%s'", change, prereq)
+                        if prereq not in self.change_success and prereq not in self.change_skipped:
+
+                            # If we're in backout mode, a prereq in 'backout_ok' state
+                            # is treated as complete.
+                            if prereq.state() in [ 'backout_ok', ]:
+                                continue
+
+                            # If a prereq has failed, doesn't need to retry, and is
+                            # marked as 'onfail:continue', then we treat it as if
+                            # this prereq has been met.
+                            elif prereq.state() not in [ 'pending', 'retry' ] and prereq.on_fail_continue:
+                                log.debug("prereq failed, but is marked onfail:continue.")
+                                continue
+
+                            log.debug("pre-req '%s' not complete yet.", prereq)
+                            all_prereqs = False
+                            break
+                        pass
+
+                    if all_prereqs:
+                        log.debug("All pre-reqs for '%s' have completed. Adding to execution queue.", change)
+                        changelist.append(change)
+                    else:
+                        log.debug("change '%s' has pending pre-reqs", change)
 
         log.debug("Returning changelist: %s", changelist)
         return changelist
@@ -587,38 +708,45 @@ class ConfigLoader:
         """
         Move a change from pending to complete.
         """
-        if change.state == CHANGE_STATE['success']:
+        log.debug("marking change as complete.")
+        if change.state() == 'success':
             self.change_success.append(change)
             self.pending_changes.remove(change)
 
-        elif change.state == CHANGE_STATE['partial_failure']:
+        elif change.state() == 'skipped':
+            log.debug("skipping change. removing from pending queue: %s", self.pending_changes)
+            self.change_skipped.append(change)
+            self.pending_changes.remove(change)
+            log.debug("pending changes are now: %s", self.pending_changes)
+            
+        elif change.state() == 'partial_failure':
             self.change_failure.append(change)
             self.pending_changes.remove(change)
 
-        elif change.state == CHANGE_STATE['total_failure']:
+        elif change.state() == 'total_failure':
             self.change_failure.append(change)
             self.pending_changes.remove(change)
 
-        elif change.state == CHANGE_STATE['backout_ok']:
+        elif change.state() == 'backout_ok':
             self.change_failure.append(change)
             self.backout_success.append(change)
             self.pending_changes.remove(change)
 
-        elif change.state == CHANGE_STATE['backout_failed']:
+        elif change.state() == 'backout_failed':
             self.change_failure.append(change)
             self.backout_failure.append(change)
             self.pending_changes.remove(change)
 
-        elif change.state == CHANGE_STATE['retry_pending']:
+        elif change.state() == 'retry_pending':
             log.info("Change '%s' will be retried in '%d' seconds", change.name, change.retry_delay)
             pass
 
-        elif change.state == CHANGE_STATE['retry']:
+        elif change.state() == 'retry':
             log.info("Change '%s' will be retried", change.name)
             pass
 
-        elif change.state == CHANGE_STATE['pending']:
-            log.error("Change '%s' did not run.", change.name)
+        elif change.state() == 'pending':
+            log.info("Change '%s' did not run.", change.name)
             self.pending_changes.remove(change)
             pass
         
@@ -627,16 +755,6 @@ class ConfigLoader:
             self.pending_changes.remove(change)
             raise ValueError("change_complete() cannot handle change state: %s", change.state)
 
-    def _apply_namespace(self, string, namespace):
-        """
-        Apply a namespace to a string
-        Deprecated
-        """
-        if namespace is not None:
-            string = string % namespace
-
-        return string
-
     def process_dependencies(self):
         """
         Post-process the dependency tree after loading it.
@@ -658,6 +776,9 @@ class ChangeController:
 
         self.current_changelist = []
 
+        # a Deferred that gets set up when do_changes() is called
+        self.alldone = None 
+
     def do_changes(self, ignored):
         self.alldone = defer.Deferred()
 
@@ -672,6 +793,7 @@ class ChangeController:
         return self.alldone
 
     def get_next_changes(self, results):
+
         """
         Fetch the next set of changes to run.
         This collects all changes that don't have any
@@ -679,6 +801,14 @@ class ChangeController:
         their pre-requisites complete successfully.
         These changes can therefore all be run together, since they
         have no interdependencies on one another.
+
+        When in backout mode, we do sortof the opposite: we find
+        all the changes that have nothing that depends on them, or
+        that have already been backed out.
+
+        FIXME: This is terribly inefficient. It would be better to
+        build a Directed Acyclic Graph at the beginning, and then
+        traverse it during execution.
         """
         log.debug("Results in get_next_changes are: %s", results)
         log.debug("Fetching outstanding changes...")
@@ -686,11 +816,13 @@ class ChangeController:
 
         if len(self.current_changelist) == 0:
             if len(self.cfgldr.pending_changes) > 0:
-                log.error("Pending changes exist that cannot be executed!")
+                log.info("Pending changes exist that cannot be executed!")
                 self.alldone.errback(ValueError("Pending changes cannot be executed"))
             else:
                 # No more changes, all changes completed.
-                self.alldone.callback("All changes complete")
+                if self.alldone is not None:
+                    self.alldone.callback("All changes complete")
+                    #self.alldone = None
             pass
         else:
             # If there are pending changes, add that call to the deferred
@@ -703,6 +835,8 @@ class ChangeController:
         It runs all of them simultaneously, since they have been identified
         as changes that can be run together. If you want to run changes
         sequentially, then they should have sequential dependencies.
+
+        FIXME: Maybe add a flag to allow sequential mode operation?
         """
         log.debug("Running pending changes...")
         dlist = []
@@ -718,7 +852,8 @@ class ChangeController:
             dlist.append(d)
             pass
 
-        dl = defer.DeferredList(dlist, fireOnOneErrback=True, consumeErrors=True)
+        dl = defer.DeferredList(dlist, fireOnOneErrback=True, consumeErrors=True )
+
         # After the current crop of changes has been completed, find if there are more
         dl.addCallback( self.get_next_changes )
         dl.addErrback( self.bailout )
@@ -730,13 +865,12 @@ class ChangeController:
         and bails out.
         """
         log.debug("Running bailout because of: %s", failure)
-        tlog.err(failure)
+        #tlog.err(failure)
         if failure.value.subFailure.type == UserBailout:
             self.alldone.callback('Bailout')
-
         else:
-            log.debug("Unhandled failure: %s", failure)
-            self.alldone.errback(failure)
+            log.debug("Unhandled bailout failure: %s", failure)
+            self.alldone.errback(failure.value.subFailure)
 
     def change_complete(self, result, change):
         """
@@ -753,18 +887,21 @@ class ChangeController:
         elif failure.type == ChangeConditionFailure:
             log.debug("Controller detected ChangeConditionFailure")
             self.change_complete(failure, change)
-            pass
+            return failure
 
         elif failure.type == NoCommands:
             #log.error(failure.value)
             self.change_complete(failure, change)
             return failure
 
+        elif failure.type == NoTargetsError:
+            return failure
+        
         else:
-            log.error("Major change failure!")
+            log.error("Major change failure: %s", failure.value)
             tlog.err(failure)
             self.change_complete(failure, change)
-            pass
+            return failure
 
     def print_stats(self, ignored):
         """
@@ -772,6 +909,7 @@ class ChangeController:
         """
         success_count = len(self.cfgldr.change_success)
         failure_count = len(self.cfgldr.change_failure)
+        skip_count = len(self.cfgldr.change_skipped)
         backout_success_count = len(self.cfgldr.backout_success)
         backout_failure_count = len(self.cfgldr.backout_failure)
         pending_count = len(self.cfgldr.pending_changes)
@@ -782,9 +920,9 @@ class ChangeController:
         if not self.cfgldr.options.backout:
         
             if not (success_count > 0 or failure_count > 0):
-                log.error("no changes succeeded or failed!")
+                log.info("no changes succeeded or failed!")
             else:
-                log.info("%d ok, %d failed (%d%% success rate)" % (success_count, failure_count, 100 * success_count / (success_count+failure_count) ) )
+                log.info("%d ok, %d failed, %d skipped (%d%% success rate)" % (success_count, failure_count, skip_count, 100 * success_count / (success_count+failure_count) ) )
 
             if success_count > 0:
                 log.info( "Successful changes: %s" % ', '.join([ x.name for x in self.cfgldr.change_success]))
@@ -792,6 +930,11 @@ class ChangeController:
             if failure_count > 0:
                 log.info( "Failed changes: %s" % ', '.join([ x.name for x in self.cfgldr.change_failure]) )
                 pass
+
+            if skip_count > 0:
+                log.info( "Skipped changes: %s" % ', '.join([ x.name for x in self.cfgldr.change_skipped]) )
+                pass
+
             pass
 
         # These stats are always printed
@@ -803,7 +946,7 @@ class ChangeController:
             log.info( "Changes backed out ok: %s" % ', '.join([ x.name for x in self.cfgldr.backout_success]))
 
         if backout_failure_count > 0:
-            log.error( "Changes NOT backed out ok: %s" % ', '.join([ x.name for x in self.cfgldr.backout_failure]))
+            log.info( "Changes NOT backed out ok: %s" % ', '.join([ x.name for x in self.cfgldr.backout_failure]))
 
         if pending_count > 0:
             if pending_count > 1:
@@ -816,6 +959,11 @@ class ChangeController:
 
             log.info( "%d change%s %s not attempted: %s" % (pending_count, plural, waswere, ', '.join([ x.name for x in self.cfgldr.pending_changes ]) ) )
 
+        # We return a failure count that is the sum of
+        # all the changes that failed, but the ones
+        # that didn't run because they couldn't be.
+        # This ensure we have a useful exit code.
+        return (success_count, failure_count + pending_count)
 
 if __name__ == '__main__':
 
@@ -841,8 +989,8 @@ if __name__ == '__main__':
         pass
 
     def errstop(failure):
-        log.error("Changes not implemented ok!")
-        tlog.err(failure)
+        log.info("Changes not implemented ok!")
+        #tlog.err(failure)
         reactor.stop()
         
     def go():
