@@ -17,6 +17,7 @@ from zope.interface import Interface, implements
 
 from twisted.internet import defer, reactor
 from twisted.internet import protocol, ssl
+from twisted.internet.error import AlreadyCalled
 from twisted.python import log as tlog
 from twisted.web import http, error, client
 
@@ -30,8 +31,8 @@ log = logging.getLogger('modipy')
 from provisioner import IProvisioner, Provisioner
 from change import Change, ChangeConditionFailure, UserBailout
 from change_command import ExpectSet, CommandChange
-from zapi import ZAPITool
-import util
+from netappzapi.zapi import ZAPITool
+from modipy import util
 
 class IZAPIProvisioner(IProvisioner):
     pass
@@ -53,15 +54,20 @@ class ZAPIProvisioner(Provisioner):
     def __init__(self, name='', namespace={},
                  authoritarian=False, autobackout=False,
                  sessionlog=None,
-                 command_timeout=300):
+                 command_timeout=300,
+                 command_delay=0):
         """
         Create a new ZAPIProvisioner.
         """
         # FIXME: May not be required
         Provisioner.__init__(self, name, namespace, authoritarian, autobackout,
                              sessionlog=sessionlog)
+
         self.command_timeout = 300
 
+        # Optional delay between commands
+        self.command_delay = int(command_delay)
+
         # Set up a deferred dictionary, keyed to uniquely identify a set
         # of commands (ExpectSet) running on given device
         self.all_commands_defer = {}
@@ -73,7 +79,7 @@ class ZAPIProvisioner(Provisioner):
 
         self.zapi_tool = ZAPITool()
 
-    def zapi_request(self, ignored, device, command, timeout=30):
+    def zapi_request(self, device, command, timeout=30):
         """
         Issue a ZAPI request to a device.
         """
@@ -81,7 +87,7 @@ class ZAPIProvisioner(Provisioner):
         log.debug("issuing command: %s", command)
         return self.zapi_tool.zapi_request(device, command, timeout=timeout)
             
-    def zapi_system_command(self, ignored, device, command, timeout=30):
+    def zapi_system_command(self, device, command, timeout=30):
         """
         Run a commandline command via ZAPI.
         """
@@ -122,116 +128,17 @@ class ZAPIProvisioner(Provisioner):
         deferred_key = (device, expectset)
         self.all_commands_defer[deferred_key] = defer.Deferred()
 
-        d = defer.succeed(None)
-
         # If we have ordinary commands, use the system-cli mechanism for running them
         if len(expectset.commands) > 0:
             log.debug("Running <system-cli/> commands...")
 
-            for (expr, cmdstring) in expectset.commands:
-
-                try:
-                    # Perform variable substitution on the command string
-                    log.debug("determining commandstring from template: %s", cmdstring)
-                    cmdstring = str(util.substituteVariables(cmdstring, namespace))
-                    log.debug("commandstring is: %s", cmdstring)
-
-                    # add the cmdstring to the namespace
-                    namespace['command.send'] = cmdstring
-
-                except KeyError, e:
-                    log.error("KeyError in commands: %s" % e)
-                    self.all_commands_defer.errback( e )
-                    return self.all_commands_defer[deferred_key]
-
-                d.addCallback(self.zapi_system_command, device, cmdstring, self.command_timeout)
-                d.addCallbacks(self.command_completed, self.command_failed, callbackArgs=(namespace,), errbackArgs=(namespace,) )
-                pass
-            pass
-
+            d = self.run_systemcli_commands(device, expectset, namespace, delay)
+            
         # If we have zapi based commands, pass them directly to the zapi-request mechanism
         elif len(expectset.zapi_commands) > 0:
             log.debug("Running zapi commands: %s", expectset.zapi_commands)
 
-            for cmdxml in expectset.zapi_commands:
-
-                cmdxml = str(util.substituteVariables(cmdxml, namespace))
-                # split the xml into the actual commands
-                cmdnode = etree.fromstring( cmdxml, etree.XMLParser(ns_clean=True))
-                if cmdnode.tag == 'zapicommand':
-                    cmdstring = etree.tostring(cmdnode[0])
-                    log.debug("cmdstring: %s", cmdstring)
-                    try:
-                        # Perform variable substitution on the command string
-##                         log.debug("determining commandstring from template: %s", cmdstring)
-##                         cmdstring = str(util.substituteVariables(cmdstring, namespace))
-##                         log.debug("commandstring is: %s", cmdstring)
-
-                        # add the cmdstring to the namespace
-                        namespace['command.send'] = cmdstring
-
-                    except KeyError, e:
-                        log.error("KeyError in commands: %s" % e)
-                        self.all_commands_defer.errback( e )
-                        return self.all_commands_defer[deferred_key]
-
-                    # Do substitution on the device with the current namespace
-                    device_addr = util.substituteVariables(str(device), namespace)
-                    d.addCallback(self.zapi_request, device_addr, cmdstring, self.command_timeout)
-                    d.addCallbacks(self.command_completed, self.command_failed, callbackArgs=(namespace,), errbackArgs=(namespace,) )
-
-                elif cmdnode.tag == 'getfile':
-                    log.info("Getting file!")
-                    # get the variables we need
-                    remotepath_node = cmdnode.find('remotepath')
-                    remotepath = remotepath_node.text
-                    filer_root = remotepath_node.attrib.get('root', '/vol/vol0')
-                    
-                    localpath_node = cmdnode.find('localpath')
-                    if localpath_node is None:
-                        localpath = remotepath
-                        local_absolute = False
-                        create_pathtree = True
-                    else:
-                        localpath = localpath_node.text
-                        local_absolute = bool(localpath_node.attrib.get('absolute', False))
-                    
-                    log.debug("remotepath: %s", remotepath)
-                    log.debug("filer_root: %s", filer_root)
-                    
-                    d.addCallback(self.zapi_get_file, device, remotepath)
-                    d.addCallback(self.zapi_file_received, device, remotepath, localpath, local_absolute)
-                    d.addCallbacks(self.command_completed, self.command_failed, callbackArgs=(namespace,), errbackArgs=(namespace,) )
-                    
-                elif cmdnode.tag == 'putfile':
-                    log.info("Putting file!")
-
-                    # get the variables we need
-                    localpath_node = cmdnode.find('localpath')
-                    localpath = localpath_node.text
-
-                    remotepath_node = cmdnode.find('remotepath')
-                    remotepath = remotepath_node.text
-                    filer_root = remotepath_node.attrib.get('root', '/vol/vol0')
-                    append = remotepath_node.attrib.get('append', False)
-                    overwrite = remotepath_node.attrib.get('overwrite', True)
-                    offset = int(remotepath_node.attrib.get('offset', 0))
-                    
-                    log.debug("localpath: %s", localpath)
-                    log.debug("remotepath: %s", remotepath)
-                    log.debug("filer_root: %s", filer_root)
-                    if append:
-                        offset = -1
-                        overwrite = True
-
-                    log.debug("append: %s, offset: %s", append, offset)
-                    log.debug("overwrite: %s", overwrite)
-                    
-                    d.addCallback(self.zapi_put_file, device, localpath, remotepath, filer_root, offset, overwrite)
-                    d.addCallback(self.zapi_file_sent, device, remotepath, localpath)
-                    d.addCallbacks(self.command_completed, self.command_failed, callbackArgs=(namespace,), errbackArgs=(namespace,) )
-                    pass
-                pass
+            d = self.run_zapi_commands(device, expectset, namespace, delay)
             pass
 
         # Set up callbacks for when all commands are finished.
@@ -240,9 +147,133 @@ class ZAPIProvisioner(Provisioner):
 
         # Set up a timeout to fire if this isn't called back fast enough
         # FIXME: Use the actual timeout value here
-        self.all_commands_t[deferred_key] = reactor.callLater(30, self.all_commands_timeout, deferred_key)
+        self.all_commands_t[deferred_key] = reactor.callLater(self.command_timeout, self.all_commands_timeout, deferred_key)
         return self.all_commands_defer[deferred_key]
 
+    def run_systemcli_commands(self, device, expectset, namespace, delay=0):
+        """
+        Run NetApp commandline commands via the ZAPI
+        """
+        d = defer.succeed(None)        
+        for (expr, cmdstring) in expectset.commands:
+
+            try:
+                # Perform variable substitution on the command string
+                log.debug("determining commandstring from template: %s", cmdstring)
+                cmdstring = str(util.substituteVariables(cmdstring, namespace))
+                log.debug("commandstring is: %s", cmdstring)
+
+                # add the cmdstring to the namespace
+                namespace['command.send'] = cmdstring
+
+            except KeyError, e:
+                log.error("KeyError in commands: %s" % e)
+                self.all_commands_defer.errback( e )
+                return self.all_commands_defer[deferred_key]
+
+            device_addr = util.substituteVariables(str(device), namespace)
+            def callsoon(ignored, device_addr, cmdstring):
+                return util.run_after_delay(self.command_delay, self.zapi_system_command, device_addr, cmdstring, self.command_timeout)
+
+            d.addCallback(callsoon, device_addr, cmdstring)
+            d.addCallbacks(self.command_completed, self.command_failed, callbackArgs=(namespace,), errbackArgs=(namespace,) )
+            pass
+        return d
+
+    def run_zapi_commands(self, device, expectset, namespace, delay):
+        """
+        Run ZAPI commands
+        """
+        d = defer.succeed(None)
+        for cmdxml in expectset.zapi_commands:
+            cmdxml = str(util.substituteVariables(cmdxml, namespace))
+            # split the xml into the actual command
+            cmdnode = etree.fromstring( cmdxml, etree.XMLParser(ns_clean=True))
+
+            if cmdnode.tag == 'zapicommand':
+                cmdstring = etree.tostring(cmdnode[0])
+                log.debug("cmdstring: %s", cmdstring)
+                try:
+                    # Perform variable substitution on the command string
+                    cmdstring = str(util.substituteVariables(cmdstring, namespace))
+                    # add the cmdstring to the namespace
+                    namespace['command.send'] = cmdstring
+
+                except KeyError, e:
+                    log.error("KeyError in commands: %s" % e)
+                    self.all_commands_defer.errback( e )
+                    return self.all_commands_defer[deferred_key]
+
+                # Do substitution on the device with the current namespace
+                device_addr = util.substituteVariables(str(device), namespace)
+
+                def callsoon(ignored, device_addr, cmdstring):
+                    log.debug("Setting up callback of: %s", cmdstring)
+                    return util.run_after_delay(self.command_delay, self.zapi_request, device_addr, cmdstring, self.command_timeout)
+
+                # Use the variables that change in the loop here, or the wrong
+                # one will be in context when the callback fires. Everything
+                # static can be left inside callsoon()
+                d.addCallback(callsoon, device_addr, cmdstring)
+
+                #d.addCallback(self.zapi_request, device_addr, cmdstring, self.command_timeout)
+                d.addCallbacks(self.command_completed, self.command_failed, callbackArgs=(namespace,), errbackArgs=(namespace,) )
+
+            elif cmdnode.tag == 'getfile':
+                log.info("Getting file!")
+                # get the variables we need
+                remotepath_node = cmdnode.find('remotepath')
+                remotepath = remotepath_node.text
+                filer_root = remotepath_node.attrib.get('root', '/vol/vol0')
+
+                localpath_node = cmdnode.find('localpath')
+                if localpath_node is None:
+                    localpath = remotepath
+                    local_absolute = False
+                    create_pathtree = True
+                else:
+                    localpath = localpath_node.text
+                    local_absolute = bool(localpath_node.attrib.get('absolute', False))
+
+                log.debug("remotepath: %s", remotepath)
+                log.debug("filer_root: %s", filer_root)
+
+                d.addCallback(self.zapi_get_file, device, remotepath)
+                d.addCallback(self.zapi_file_received, device, remotepath, localpath, local_absolute)
+                d.addCallbacks(self.command_completed, self.command_failed, callbackArgs=(namespace,), errbackArgs=(namespace,) )
+
+            elif cmdnode.tag == 'putfile':
+                log.info("Putting file!")
+
+                # get the variables we need
+                localpath_node = cmdnode.find('localpath')
+                localpath = localpath_node.text
+
+                remotepath_node = cmdnode.find('remotepath')
+                remotepath = remotepath_node.text
+                filer_root = remotepath_node.attrib.get('root', '/vol/vol0')
+                append = remotepath_node.attrib.get('append', False)
+                overwrite = remotepath_node.attrib.get('overwrite', True)
+                offset = int(remotepath_node.attrib.get('offset', 0))
+
+                log.debug("localpath: %s", localpath)
+                log.debug("remotepath: %s", remotepath)
+                log.debug("filer_root: %s", filer_root)
+                if append:
+                    offset = -1
+                    overwrite = True
+
+                log.debug("append: %s, offset: %s", append, offset)
+                log.debug("overwrite: %s", overwrite)
+
+                d.addCallback(self.zapi_put_file, device, localpath, remotepath, filer_root, offset, overwrite)
+                d.addCallback(self.zapi_file_sent, device, remotepath, localpath)
+                d.addCallbacks(self.command_completed, self.command_failed, callbackArgs=(namespace,), errbackArgs=(namespace,) )
+                pass
+            pass
+
+        return d
+    
     def zapi_file_received(self, (results, filedata), device, remotepath, localpath, absolute=False):
         """
         File data is received for a remote file. Save it locally in localpath
@@ -298,14 +329,20 @@ class ZAPIProvisioner(Provisioner):
         Called when all the commands have completed.
         """
         log.debug("All commands have finished.")
-        self.all_commands_t[deferred_key].cancel()
+        try:
+            self.all_commands_t[deferred_key].cancel()
+        except AlreadyCalled:
+            pass
         self.all_commands_defer[deferred_key].callback( result )
         
     def all_commands_failure(self, failure, deferred_key, namespace):
         """
         Some kind of failure in the commands occurred.
         """
-        self.all_commands_t[deferred_key].cancel()
+        try:
+            self.all_commands_t[deferred_key].cancel()
+        except AlreadyCalled:
+            pass
         self.all_commands_defer[deferred_key].errback( failure )
 
     def all_commands_timeout(self, deferred_key):
