@@ -1,5 +1,5 @@
 #
-# $Id: provisioner_command.py 98 2009-07-04 07:35:19Z daedalus $
+# $Id: provisioner_command.py 99 2009-08-25 22:15:25Z daedalus $
 #
 ##COPYRIGHT##
 
@@ -320,7 +320,7 @@ class ExpectProtocol(protocol.Protocol):
         Generic function used when data is received from the child process,
         whatever kind of subprocess it is.
         """
-        #log.debug("data: %s" % data)
+        log.debug("data: %s" % data)
         self.databuf += data
         if self.sessionlog is not None:
             self.sessionlog.write(data)
@@ -507,12 +507,20 @@ class MultiConnectingProvisioner(ConnectingProvisioner):
     ssh remotehost <command1>
     ssh remotehost <command2>
     """
+    required_tags = ConnectingProvisioner.required_tags
+    required_tags.remove('connect_command')
+    
+    def __init__(self, name, namespace={},
+                 authoritarian=False, autobackout=False,
+                 sessionlog=None,
+                 command_timeout=300):
 
-    def __init__(self, name='', namespace={}, authoritarian=False, autobackout=False, command_timeout=300):
+        self.command = ''
 
         ConnectingProvisioner.__init__(self, name, namespace, authoritarian, autobackout, command_timeout)
         # Set up a deferred dictionary, keyed to uniquely identify a set
         # of commands (ExpectSet) running on given device
+
         self.all_commands_defer = {}
 
     def parse_config_node(self, node):
@@ -521,7 +529,7 @@ class MultiConnectingProvisioner(ConnectingProvisioner):
         """
         ConnectingProvisioner.parse_config_node(self, node)
         # parse command definition
-        if node.tag == 'command':
+        if node.tag == 'connect_command':
             self.command = str(node.text)
             log.debug("Set provisioner command to: %s", self.command)
 
@@ -565,25 +573,27 @@ class MultiConnectingProvisioner(ConnectingProvisioner):
         for (expr, cmdstring, timeout) in expectset.commands:
 
             try:
+                # This only runs a single command at a time,
+                # which is a combination of the 'connect' command
+                # (usually an ssh or sudo or something similar)
+                # followed by the command we want to run remotely
+
+                command = ' '.join([self.command, cmdstring])
+                
                 # Perform variable substitution on the command string
-                log.debug("determining commandstring from template: %s", cmdstring)
-                cmdstring = str(util.substituteVariables(cmdstring, namespace))
-                log.debug("commandstring is: %s", cmdstring)
+                command = str(util.substituteVariables(command, namespace))
+                log.debug("commandstring is: %s", command)
 
                 # add the cmdstring to the namespace
-                namespace['command.send'] = cmdstring
-
-                log.debug("Determining command base from template: %s", self.command)
-                command = str(util.substituteVariables(self.command, namespace))
+                namespace['command.send'] = command
 
             except KeyError, e:
                 log.error("KeyError in commands: %s" % e)
                 self.all_commands_defer.errback( e )
                 return self.all_commands_defer[deferred_key]
             
-            #log.debug("checking expr: %s" % expr)
             log.debug("cmdstring is: %s" % command)
-            if command is not None:
+            if command is not None and len(command) > 0:
                 log.debug("splitting command: %s", command)
                 cmd = self.split_command(command)
                 log.debug("command is: %s", cmd)
@@ -678,18 +688,18 @@ class SingleCommandProtocol(protocol.Protocol):
         self.timeout_after = timeout_after
 
     def connectionMade(self):
-        #log.debug("Pretend connection made! Hurray!")
+        log.debug("Pretend connection made! Hurray!")
 
         # FIXME: provide a way to set this via the config file
         self.timeout = reactor.callLater(self.timeout_after, self.timedOut)
         pass
     
     def childConnectionLost(self, fd):
-        #log.debug("Lost connection to child process fd %s" % fd)
+        log.debug("Lost connection to child process fd %s" % fd)
         pass
 
     def childDataReceived(self, fd, data):
-        #log.debug("received data from child process %s: >%s<" % (fd, data) )
+        log.debug("received data from child process %s: >%s<" % (fd, data) )
         self.databuf += data
         #log.debug("databuf: >%s<" % self.databuf)
 
